package geniusweb.protocol.session;

import static org.junit.Assert.assertEquals;

import java.io.IOException;
import java.net.URISyntaxException;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.junit.Before;
import org.junit.Test;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;

import geniusweb.issuevalue.Bid;
import geniusweb.issuevalue.DiscreteValue;
import geniusweb.issuevalue.Value;
import geniusweb.protocol.SessionResult;
import geniusweb.references.PartyRef;
import geniusweb.references.PartyWithProfile;
import geniusweb.references.ProfileRef;
import tudelft.utilities.junit.GeneralTests;

public class SessionResultTest extends GeneralTests<SessionResult> {
	private final ObjectMapper jackson = new ObjectMapper();
	private final RuntimeException error = new RuntimeException("test");

	private final String ISSUE1 = "issue1";
	// private final String errorstring =
	// "\"error\":{\"java.lang.RuntimeException\":{\"cause\":null,\"stackTrace\":[{\"methodName\":\"<init>\",\"fileName\":\"SessionResultTest.java\",\"lineNumber\":29,\"className\":\"events.SessionResultTest\",\"nativeMethod\":false},{\"methodName\":\"newInstance0\",\"fileName\":\"NativeConstructorAccessorImpl.java\",\"lineNumber\":-2,\"className\":\"sun.reflect.NativeConstructorAccessorImpl\",\"nativeMethod\":true},{\"methodName\":\"newInstance\",\"fileName\":\"NativeConstructorAccessorImpl.java\",\"lineNumber\":62,\"className\":\"sun.reflect.NativeConstructorAccessorImpl\",\"nativeMethod\":false},{\"methodName\":\"newInstance\",\"fileName\":\"DelegatingConstructorAccessorImpl.java\",\"lineNumber\":45,\"className\":\"sun.reflect.DelegatingConstructorAccessorImpl\",\"nativeMethod\":false},{\"methodName\":\"newInstance\",\"fileName\":\"Constructor.java\",\"lineNumber\":423,\"className\":\"java.lang.reflect.Constructor\",\"nativeMethod\":false},{\"methodName\":\"createTest\",\"fileName\":\"BlockJUnit4ClassRunner.java\",\"lineNumber\":217,\"className\":\"org.junit.runners.BlockJUnit4ClassRunner\",\"nativeMethod\":false},{\"methodName\":\"runReflectiveCall\",\"fileName\":\"BlockJUnit4ClassRunner.java\",\"lineNumber\":266,\"className\":\"org.junit.runners.BlockJUnit4ClassRunner$1\",\"nativeMethod\":false},{\"methodName\":\"run\",\"fileName\":\"ReflectiveCallable.java\",\"lineNumber\":12,\"className\":\"org.junit.internal.runners.model.ReflectiveCallable\",\"nativeMethod\":false},{\"methodName\":\"methodBlock\",\"fileName\":\"BlockJUnit4ClassRunner.java\",\"lineNumber\":263,\"className\":\"org.junit.runners.BlockJUnit4ClassRunner\",\"nativeMethod\":false},{\"methodName\":\"runChild\",\"fileName\":\"BlockJUnit4ClassRunner.java\",\"lineNumber\":78,\"className\":\"org.junit.runners.BlockJUnit4ClassRunner\",\"nativeMethod\":false},{\"methodName\":\"runChild\",\"fileName\":\"BlockJUnit4ClassRunner.java\",\"lineNumber\":57,\"className\":\"org.junit.runners.BlockJUnit4ClassRunner\",\"nativeMethod\":false},{\"methodName\":\"run\",\"fileName\":\"ParentRunner.java\",\"lineNumber\":290,\"className\":\"org.junit.runners.ParentRunner$3\",\"nativeMethod\":false},{\"methodName\":\"schedule\",\"fileName\":\"ParentRunner.java\",\"lineNumber\":71,\"className\":\"org.junit.runners.ParentRunner$1\",\"nativeMethod\":false},{\"methodName\":\"runChildren\",\"fileName\":\"ParentRunner.java\",\"lineNumber\":288,\"className\":\"org.junit.runners.ParentRunner\",\"nativeMethod\":false},{\"methodName\":\"access$000\",\"fileName\":\"ParentRunner.java\",\"lineNumber\":58,\"className\":\"org.junit.runners.ParentRunner\",\"nativeMethod\":false},{\"methodName\":\"evaluate\",\"fileName\":\"ParentRunner.java\",\"lineNumber\":268,\"className\":\"org.junit.runners.ParentRunner$2\",\"nativeMethod\":false},{\"methodName\":\"run\",\"fileName\":\"ParentRunner.java\",\"lineNumber\":363,\"className\":\"org.junit.runners.ParentRunner\",\"nativeMethod\":false},{\"methodName\":\"run\",\"fileName\":\"JUnit4TestReference.java\",\"lineNumber\":86,\"className\":\"org.eclipse.jdt.internal.junit4.runner.JUnit4TestReference\",\"nativeMethod\":false},{\"methodName\":\"run\",\"fileName\":\"TestExecution.java\",\"lineNumber\":38,\"className\":\"org.eclipse.jdt.internal.junit.runner.TestExecution\",\"nativeMethod\":false},{\"methodName\":\"runTests\",\"fileName\":\"RemoteTestRunner.java\",\"lineNumber\":538,\"className\":\"org.eclipse.jdt.internal.junit.runner.RemoteTestRunner\",\"nativeMethod\":false},{\"methodName\":\"runTests\",\"fileName\":\"RemoteTestRunner.java\",\"lineNumber\":760,\"className\":\"org.eclipse.jdt.internal.junit.runner.RemoteTestRunner\",\"nativeMethod\":false},{\"methodName\":\"run\",\"fileName\":\"RemoteTestRunner.java\",\"lineNumber\":460,\"className\":\"org.eclipse.jdt.internal.junit.runner.RemoteTestRunner\",\"nativeMethod\":false},{\"methodName\":\"main\",\"fileName\":\"RemoteTestRunner.java\",\"lineNumber\":206,\"className\":\"org.eclipse.jdt.internal.junit.runner.RemoteTestRunner\",\"nativeMethod\":false}],\"message\":\"test\",\"localizedMessage\":\"test\",\"suppressed\":[]}}";
	private SessionResult result1, result1a, result2, result3, result4;
	private String errorstring; // created dynamically: eclipse and maven
								// generate different stacktrace.
	private String jsonstring;

	@Before
	public void before() throws URISyntaxException, JsonProcessingException {
		errorstring = "\"error\":{\"java.lang.RuntimeException\":"
				+ jackson.writeValueAsString(error) + "}";
		System.out.println(errorstring);
		jsonstring = "{\"participants\":[{\"party\":\"party1\",\"profile\":\"profile1\"},{\"party\":\"party2\",\"profile\":\"profile2\"}],\"agreement\":{\"issuevalues\":{\"issue1\":\"a\"}},"
				+ errorstring + "}";

		PartyWithProfile partyprofile1 = new PartyWithProfile(
				new PartyRef("party1"), new ProfileRef("profile1"));
		PartyWithProfile partyprofile2 = new PartyWithProfile(
				new PartyRef("party2"), new ProfileRef("profile2"));

		Map<String, Value> issuevalues1 = new HashMap<>();
		issuevalues1.put(ISSUE1, new DiscreteValue("a"));
		Bid bid1 = new Bid(issuevalues1);

		// different order but that shouldn't matter
		Map<String, Value> issuevalues2 = new HashMap<>();
		issuevalues2.put(ISSUE1, new DiscreteValue("b"));
		Bid bid2 = new Bid(issuevalues2);

		result1 = new SessionResult(Arrays.asList(partyprofile1, partyprofile2),
				bid1, error);
		result1a = new SessionResult(
				Arrays.asList(partyprofile1, partyprofile2), bid1, error);
		result2 = new SessionResult(Arrays.asList(partyprofile1, partyprofile2),
				bid2, error);
		result3 = new SessionResult(Arrays.asList(partyprofile2, partyprofile1),
				bid2, error);
		result4 = new SessionResult(Arrays.asList(partyprofile1, partyprofile2),
				bid1, null);
	}

	@Override
	public List<List<SessionResult>> getGeneralTestData() {
		return Arrays.asList(Arrays.asList(result1, result1a),
				Arrays.asList(result2), Arrays.asList(result3),
				Arrays.asList(result4));
	}

	@Override
	public List<String> getGeneralTestStrings() {
		return Arrays.asList(
				"SessionResult.*party1.*profile1.*,.*party2.*profile2.*Bid.*issue1=\"a\".*java.lang.RuntimeException.*test.*",
				"SessionResult.*party1.*profile1.*,.*party2.*profile2.*Bid.*issue1=\"b\".*java.lang.RuntimeException.*test.*",
				"SessionResult.*party2.*profile2.*,.*party1.*profile1.*Bid.*issue1=\"b\".*java.lang.RuntimeException.*test.*",
				"SessionResult.*party1.*profile1.*,.*party2.*profile2.*Bid.*issue1=\"a\".*null.*");
	}

	@Test
	public void serializeTest() throws JsonProcessingException {
		System.out.println(jackson.writeValueAsString(result1));
		assertEquals(jsonstring, jackson.writeValueAsString(result1));
	}

	@Test
	public void deserializeTest() throws IOException {
		SessionResult act = jackson.readValue(jsonstring, SessionResult.class);
		assertEquals(result1, act);
	}

}
