package geniusweb.protocol.session.saop;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

import java.util.Arrays;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;

import org.junit.Before;
import org.junit.Test;

import geniusweb.actions.Accept;
import geniusweb.actions.Action;
import geniusweb.actions.EndNegotiation;
import geniusweb.actions.Offer;
import geniusweb.actions.PartyId;
import geniusweb.issuevalue.Bid;
import geniusweb.progress.Progress;
import geniusweb.progress.ProgressRounds;
import geniusweb.progress.ProgressTime;
import geniusweb.protocol.ProtocolException;
import geniusweb.protocol.partyconnection.ConnectionWithParties;
import geniusweb.protocol.partyconnection.ConnectionWithParty;
import geniusweb.protocol.session.SessionSettings;
import tudelft.utilities.junit.GeneralTests;

public class SAOPStateTest extends GeneralTests<SAOPState> {
	private final long NOW = 1000;

	private PartyId party1 = new PartyId("party1");
	private PartyId party2 = new PartyId("party2");
	private PartyId party3 = new PartyId("party3");
	private List<Action> actions = new LinkedList<>();
	private Action action = mock(Action.class);
	private List<Action> actions2 = Arrays.asList(action);

	private ConnectionWithParty party1conn = mock(ConnectionWithParty.class);
	private ConnectionWithParty party2conn = mock(ConnectionWithParty.class);
	private ConnectionWithParty party3conn = mock(ConnectionWithParty.class);

	private ConnectionWithParties connections = new ConnectionWithParties(
			Arrays.asList(party1conn, party2conn));
	private ConnectionWithParties connections2 = new ConnectionWithParties(
			Arrays.asList(party1conn, party3conn));
	private ConnectionWithParties connections3 = new ConnectionWithParties(
			Arrays.asList(party1conn, party2conn, party3conn));

	private Progress progresstime = mock(ProgressTime.class);
	private ProgressRounds progressrounds = mock(ProgressRounds.class);
	private ProgressRounds progressrounds1 = mock(ProgressRounds.class);
	private SessionSettings settings = mock(SAOPSettings.class);
	private SAOPState state1, state1a, state2, state3, state4;

	private Bid bid1 = mock(Bid.class);
	private Bid otherbid = mock(Bid.class);
	private Accept accept1 = new Accept(party1, bid1);
	private Accept accept2 = new Accept(party2, bid1);
	private Accept acceptother = new Accept(party2, otherbid);
	private Offer offer1 = new Offer(party1, bid1);
	private EndNegotiation endNegotiation1 = new EndNegotiation(party1);

	@Before
	public void before() {
		when(party1conn.getParty()).thenReturn(party1);
		when(party2conn.getParty()).thenReturn(party2);
		when(party3conn.getParty()).thenReturn(party3);
		when(party1conn.toString()).thenReturn("conn1");
		when(party2conn.toString()).thenReturn("conn2");
		when(party3conn.toString()).thenReturn("conn3");
		when(action.toString()).thenReturn("act1");

		when(progressrounds.getTerminationTime())
				.thenReturn(new Date(System.currentTimeMillis() + 3600000l));
		when(progressrounds1.getTerminationTime())
				.thenReturn(new Date(System.currentTimeMillis() + 3600000l));
		when(progressrounds.advance()).thenReturn(progressrounds1);

		state1 = new SAOPState(actions, connections, progresstime, settings,
				null, null);
		state1a = new SAOPState(actions, connections, progresstime, settings,
				null, null);
		state2 = new SAOPState(actions2, connections, progresstime, settings,
				null, null);
		state3 = new SAOPState(actions, connections2, progresstime, settings,
				null, null);
		state4 = new SAOPState(actions, connections, progressrounds, settings,
				null, null);

	}

	@Override
	public List<List<SAOPState>> getGeneralTestData() {
		return Arrays.asList(Arrays.asList(state1, state1a),
				Arrays.asList(state2), Arrays.asList(state3),
				Arrays.asList(state4));
	}

	@Override
	public List<String> getGeneralTestStrings() {
		return Arrays.asList(
				"SAOPState.*\\[\\].*ConnectionWithParties.*conn1.*conn2.*ProgressTime.*",
				"SAOPState.*\\[act1\\].*ConnectionWithParties.*conn1.*conn2.*ProgressTime.*",
				"SAOPState.*ConnectionWithParties.*conn1.*conn3.*ProgressTime.*",
				"SAOPState.*ConnectionWithParties.*conn1.*conn2.*ProgressRounds.*");
	}

	@Test
	public void constructWith0Connection() {
		new SAOPState(actions, new ConnectionWithParties(Arrays.asList()),
				progresstime, settings, null, null);
	}

	@Test
	public void constructWithNullConnection() {
		SAOPState state = new SAOPState(actions, null, progresstime, settings,
				null, null);
		assertEquals(0, state.getConnections().size());
	}

	@Test
	public void testNextActor() {
		assertEquals(party1, state1.getNextActor());
	}

	@Test
	public void testNextActor2() {
		ConnectionWithParties conns = new ConnectionWithParties(
				Arrays.asList(party2conn, party3conn));
		SAOPState state = new SAOPState(actions, conns, progresstime, settings,
				null, null);
		assertEquals(party2, state.getNextActor());
	}

	@Test
	public void testNextActorAfterAction() {
		List<Action> actions = Arrays.asList(action);
		SAOPState state = new SAOPState(actions, connections, progresstime,
				settings, null, null);
		assertEquals(party2, state.getNextActor());
	}

	@Test
	public void testNextActorAfter2Actions() {
		List<Action> actions = Arrays.asList(action, action);
		SAOPState state = new SAOPState(actions, connections, progresstime,
				settings, null, null);
		assertEquals(party1, state.getNextActor());
	}

	@Test
	public void isFinalTest() {
		assertFalse(state1.isFinal(NOW));
		SAOPState state = new SAOPState(actions, connections, progresstime,
				settings, null, mock(ProtocolException.class));
		assertTrue(state.isFinal(NOW));
	}

	@Test
	public void withExceptionTest() {
		SAOPState finalstate1 = state1.with(mock(ProtocolException.class));
		assertTrue(finalstate1.isFinal(NOW));
		SAOPState finalstate2 = finalstate1.with(mock(ProtocolException.class));
		assertTrue(finalstate2.isFinal(NOW));
	}

	@Test
	public void isOfferAgreementTest1() {
		List<Action> actions = Arrays.asList(offer1);
		SAOPState state = new SAOPState(actions, connections3, progressrounds,
				settings, null, null);
		assertNull(state.getAgreement());
	}

	@Test
	public void isOfferAcceptAgreementTest() {
		List<Action> actions = Arrays.asList(offer1, accept2);
		SAOPState state = new SAOPState(actions, connections3, progressrounds,
				settings, null, null);
		assertNull(state.getAgreement());
	}

	@Test
	public void isOfferAcceptOfferAgreementTest() {
		List<Action> actions = Arrays.asList(offer1, accept1, offer1);
		SAOPState state = new SAOPState(actions, connections3, progressrounds,
				settings, null, null);
		assertNull(state.getAgreement());
	}

	@Test
	public void isOfferAcceptAcceptAgreementTest() {
		List<Action> actions = Arrays.asList(offer1, accept1, accept2);
		SAOPState state = new SAOPState(actions, connections3, progressrounds,
				settings, null, null);
		assertEquals(bid1, state.getAgreement());
		assertTrue(state.isFinal(NOW));
	}

	@Test
	public void isAcceptOfferAcceptAgreementTest() {
		List<Action> actions = Arrays.asList(accept1, offer1, accept2);
		SAOPState state = new SAOPState(actions, connections3, progressrounds,
				settings, null, null);
		assertEquals(null, state.getAgreement());
	}

	@Test
	public void isOfferAcceptAcceptotherAgreementTest() {
		List<Action> actions = Arrays.asList(offer1, accept1, acceptother);
		SAOPState state = new SAOPState(actions, connections3, progressrounds,
				settings, null, null);
		assertEquals(null, state.getAgreement());
	}

	@Test
	public void withWrongActionTest() {
		// wrong because action has actor null which does not match party1
		SAOPState state = state1.with(party1, action);
		assertEquals(1, state.getActions().size());
		assertEquals(action, state.getActions().get(0));
		assertTrue(state.isFinal(NOW));
		assertEquals(
				"geniusweb.protocol.ProtocolException: party1:act contains wrong credentials: act1",
				state.getError().toString());
	}

	@Test
	public void withActionTest() {
		when(action.getActor()).thenReturn(party1);
		SAOPState state = state1.with(party1, action);
		assertEquals(1, state.getActions().size());
		assertEquals(action, state.getActions().get(0));
	}

	@Test
	public void withConnectionTest() {
		SAOPState state = state1.with(party3conn, null);
		assertEquals(3, state.getConnections().size());
	}

	@Test
	public void isEndNegotiationFinalTest() {
		List<Action> actions = Arrays.asList(endNegotiation1);
		SAOPState state = new SAOPState(actions, connections3, progressrounds,
				settings, null, null);
		assertEquals(null, state.getAgreement());
		assertEquals(null, state.getError());
		assertTrue(state.isFinal(NOW));

	}

	@Test
	public void isProtocolErrorFinal() {
		SAOPState state = new SAOPState(actions, connections3, progressrounds,
				settings, null, new ProtocolException("test error", "test"));
		assertTrue(state.isFinal(NOW));
		assertNotNull(state.getError());
	}

	@Test
	public void RefuseImmediateAccept() {
		Accept nullaccept = new Accept(party1, null);
		SAOPState state = state1.with(party1, nullaccept);
		assertTrue(state.isFinal(NOW));
		assertEquals(
				"geniusweb.protocol.ProtocolException: party1:Accept without a recent offer",
				state.getError().toString());
	}

	@Test
	public void RefuseNotMyTurn() {
		List<Action> actions = Arrays.asList(offer1);
		SAOPState state = new SAOPState(actions, connections3, progressrounds,
				settings, null, null);

		state = state.with(party1, offer1);
		assertTrue(state.isFinal(NOW));
		assertEquals(
				"geniusweb.protocol.ProtocolException: party1:actor does not have the turn ",
				state.getError().toString());
	}

	@Test
	public void RefuseNullAccept() {
		List<Action> actions = Arrays.asList(offer1);
		SAOPState state = new SAOPState(actions, connections3, progressrounds,
				settings, null, null);

		Accept nullaccept = new Accept(party2, null);
		state = state.with(party2, nullaccept);
		assertTrue(state.isFinal(NOW));
		assertTrue(state.getError().toString()
				.matches(".*party2.*Accepts a bid that was not offered.*"));
	}

	@Test
	public void RefuseNullAction() {
		List<Action> actions = Arrays.asList(offer1);
		SAOPState state = new SAOPState(actions, connections3, progressrounds,
				settings, null, null);

		state = state.with(party2, null);
		assertEquals(
				"geniusweb.protocol.ProtocolException: party2:action is null",
				state.getError().toString());
	}

	@Test
	public void checkGoodAccept() {
		List<Action> actions = Arrays.asList(offer1, accept2);
		SAOPState state = new SAOPState(actions, connections3, progressrounds,
				settings, null, null);

		Accept accept = new Accept(party3, bid1);
		state = state.with(party3, accept);
		assertTrue(state.isFinal(NOW));
		assertNull(state.getError());
		assertEquals(bid1, state.getAgreement());

	}

	@Test
	public void checkGoodSecondAccept() {
		Bid bid2 = mock(Bid.class);
		List<Action> actions = Arrays.asList(offer1, new Offer(party2, bid2),
				new Accept(party3, bid2));
		SAOPState state = new SAOPState(actions, connections3, progressrounds,
				settings, null, null);

		Accept accept = new Accept(party1, bid2);
		state = state.with(party1, accept);
		assertTrue(state.isFinal(NOW));
		assertNull(state.getError());
		assertEquals(bid2, state.getAgreement());

	}
}
